# -*- coding: utf-8 -*-
from odoo.upgrade import util


def migrate(cr, version):
    util.create_column(cr, "sale_order", "maintenance_subscription_id", "int4")
    util.create_column(cr, "sale_order", "enterprise_saas", "boolean")
    util.create_column(cr, "sale_order", "to_renew", "boolean")
    util.create_column(cr, "sale_order", "exclude_dashboard", "boolean")
    util.create_column(cr, "sale_order", "is_staff_restricted", "boolean")
    util.create_column(cr, "sale_order", "referrer_id", "int4")
    util.create_column(cr, "sale_order", "enterprise_max_db", "int4")
    util.create_column(cr, "sale_order", "responsible_id", "int4")
    util.create_column(cr, "sale_order", "enterprise_security_email", "varchar")
    util.create_column(cr, "sale_order", "odoo_sh_extra_gb", "int4")
    util.create_column(cr, "sale_order", "odoo_sh_extra_stagings", "int4")
    util.create_column(cr, "sale_order", "odoo_sh_extra_workers", "int4")
    util.create_column(cr, "sale_order", "odoo_sh_extra_dedicated", "int4")
    util.create_column(cr, "sale_order", "pricelist_id", "int4")
    util.create_column(cr, "sale_order", "hosting", "varchar")
    util.create_column(cr, "sale_order", "internal_note", "text")
    util.create_column(cr, "sale_order", "upsell_order_id", "int4")
    util.create_column(cr, "sale_order", "upsell_date", "date")

    util.create_column(cr, "sale_order_line", "computation", "varchar")
    util.create_column(cr, "sale_order_line", "database_id", "int4")

    cr.execute("""
        UPDATE sale_order so
           SET enterprise_saas=ss.enterprise_saas
              ,payment_exception=ss.enterprise_exception
              ,to_renew=ss.to_renew
              ,exclude_dashboard=ss.exclude_dashboard
              ,is_staff_restricted=ss.is_staff_restricted
              ,referrer_id=ss.referrer_id
              ,enterprise_max_db=ss.enterprise_max_db
              ,responsible_id=ss.responsible_id
              ,enterprise_security_email=ss.enterprise_security_email
              ,odoo_sh_extra_gb=ss.odoo_sh_extra_gb
              ,odoo_sh_extra_stagings=ss.odoo_sh_extra_stagings
              ,odoo_sh_extra_workers=ss.odoo_sh_extra_workers
              ,odoo_sh_extra_dedicated=ss.odoo_sh_extra_dedicated
              ,pricelist_id=ss.pricelist_id
              ,hosting=ss.hosting
              ,internal_note=ss.notes
              ,upsell_order_id=ss.upsell_order_id
              ,upsell_date=ss.upsell_date
              ,maintenance_subscription_id = ss.maintenance_subscription_id
          FROM sale_subscription ss
         WHERE ss.id=so.old_subscription_id
    """)
    cr.execute("""
        UPDATE sale_order_line sol
           SET computation=ssl.computation
              ,database_id=ssl.database_id
          FROM sale_subscription_line ssl
         WHERE sol.old_subscription_line_id=ssl.id
    """)
    util.force_noupdate(cr, "openerp_enterprise.cron_yearly_upsell_ir_actions_server", True)

    util.create_column(cr, "sale_order_template", "oe_apps_uom_id", "int4")
    util.create_column(cr, "sale_order_template", "oe_users_uom_id", "int4")
    util.create_column(cr, "sale_order_template", "oe_mnt_uom_id", "int4")
    util.create_column(cr, "sale_order_template", "recurrence_id", "int4")

    # clean the xmlids on the code too

    # For odoo.com we use a direct link between sale_order_template and sale_subscription_template.
    # Meaning, each sale_order_template is linked to a sale_subscription_template
    # The way the odoo migration work is naïve, it is just migrated a subscription template to a sale order template.
    # Here, for odoo.com, we want the subscription to be set to their template sale order counterpart.
    # The code here does it and should be easily understandable
    xml_ids_to_migrate_map = [
        ('subscription_template_odoo_partnership', 'quotation_template_odoo_partnership'),
        ('subscription_template_odoo_monthly', 'quotation_template_odoo_monthly'),
        ('subscription_template_odoo_yearly', 'quotation_template_odoo_yearly'),
        ('subscription_template_odoo_2_years', 'quotation_template_odoo_2_years'),
        ('subscription_template_odoo_3_years', 'quotation_template_odoo_3_years'),
        ('subscription_template_odoo_4_years', 'quotation_template_odoo_4_years'),
        ('subscription_template_odoo_5_years', 'quotation_template_odoo_5_years'),
    ]

    for old, new in xml_ids_to_migrate_map:
        subscription_tmpl_id = util.ref(cr, f"openerp_enterprise.{old}")
        order_tmpl_id = util.ref(cr, f"openerp_enterprise.{new}")
        # copy OE fields values from subscription templates to SO templates
        cr.execute("""
        UPDATE
            sale_order_template sot
        SET
            oe_apps_uom_id = upgraded_sot.oe_apps_uom_id,
            oe_users_uom_id = upgraded_sot.oe_users_uom_id,
            oe_mnt_uom_id = upgraded_sot.oe_mnt_uom_id,
            recurrence_id = upgraded_sot.recurrence_id
        FROM
            sale_subscription_template sst, sale_order_template upgraded_sot
        WHERE
            sot.id = %s
            AND upgraded_sot.id = %s
            AND sst.id = upgraded_sot.old_template_id
        """, [order_tmpl_id, subscription_tmpl_id])

        # update the subscriptions' template id
        cr.execute("""
            UPDATE
                sale_order so
            SET
                sale_order_template_id=%s
            WHERE
                so.sale_order_template_id = %s
        """, [order_tmpl_id, subscription_tmpl_id])



    # Delete the migrated subscription template, as we have directly matched them with their sale order template counterpart
    # We do not need them and it will clean our prod database
    ids_to_delete = [util.ref(cr, f'openerp_enterprise.{x[0]}') for x in xml_ids_to_migrate_map]

    cr.execute("""
        DELETE FROM
            sale_order_template
        WHERE
            id = ANY(%s)
    """, [ids_to_delete])

    # Force the update for subscription email templates
    util.update_record_from_xml(cr, "openerp_enterprise.email_subscription_open", reset_translations=['body_html', 'subject'])
    util.update_record_from_xml(cr, "openerp_enterprise.subscription_open_partnership", reset_translations=['body_html', 'subject'])


    # The product_pricing table migration is quite naive and doesn't take in account the prices set on the pricelists
    # Even though we have access to the pricelist, currency, etc ... (would be very complicated to solve this use case)
    # Since we are entirely relying on pricelists and not on the prices set on the products, we need to correctly compute them.
    # He we are mapping the uom to their corresponding recurrency (yearly, monthly) and duration (from 1 month to 5 years)
    uom_mapping = [
        ('openerp_enterprise.uom_categ_month_period', 1, 'month', 'openerp_enterprise.product_uom_month'),
        ('openerp_enterprise.uom_categ_month_period', 1, 'year', 'openerp_enterprise.product_uom_year'),
        ('openerp_enterprise.uom_categ_month_period', 2, 'year', 'openerp_enterprise.product_uom_2_year'),
        ('openerp_enterprise.uom_categ_month_period', 3, 'year', 'openerp_enterprise.product_uom_3_year'),
        ('openerp_enterprise.uom_categ_month_period', 4, 'year', 'openerp_enterprise.product_uom_4_year'),
        ('openerp_enterprise.uom_categ_month_period', 5, 'year', 'openerp_enterprise.product_uom_5_year'),
        ('openerp_enterprise.uom_categ_online_user', 1, 'month', 'openerp_enterprise.product_online_user_month'),
        ('openerp_enterprise.uom_categ_online_user', 1, 'year', 'openerp_enterprise.product_online_user_year'),
        ('openerp_enterprise.uom_categ_online_user', 2, 'year', 'openerp_enterprise.product_online_user_2year'),
        ('openerp_enterprise.uom_categ_online_user', 3, 'year', 'openerp_enterprise.product_online_user_3year'),
        ('openerp_enterprise.uom_categ_online_user', 4, 'year', 'openerp_enterprise.product_online_user_4year'),
        ('openerp_enterprise.uom_categ_online_user', 5, 'year', 'openerp_enterprise.product_online_user_5year'),
    ]

    # Here we build a table to map the uom and their recurrency to be able to do all the computation in SQL instead of python code

    uom_mapping_sql = list()
    for categ_xmlid, duration, unit, uom_xml_id in uom_mapping:
        categ_id = util.ref(cr, categ_xmlid)
        uom_id = util.ref(cr, uom_xml_id)
        uom_mapping_sql.append(f'({categ_id}, {duration}, \'{unit}\', {uom_id})')

    cr.execute("""
        CREATE TABLE __oe_mig_15_3_recurrence_uom_mapping (
            category_id int4,
            duration int4,
            unit varchar,
            uom_id int4
        );
    """)

    uom_mapping_sql = ', '.join(uom_mapping_sql)
    cr.execute("INSERT INTO __oe_mig_15_3_recurrence_uom_mapping (category_id, duration, unit, uom_id) VALUES %s" % uom_mapping_sql)

    # here is done the computation.
    # The logic is explained below:

    # The product.pricing records created by the migration naively copy the product price without looking at pricelists, etc ...
    # The code here will compute the correct price for this product.pricing.
    # To do so, it will first loop through all the pricing, fetch the product/template linked to it, from there it will retrieve the
    # uom configured on the product to be able to get the uom category.
    # It will then fetch the product price set on the pricelist linked to the product pricing, it is always set for monthly recurrence.
    # From there we now retrieve the target uom on the mapping table `__oe_mig_15_3_recurrence_uom_mapping` by searching through,
    # the uom category that match the product uom's category, the recurrence (duration and unit) set on the product_pricing.
    # With the target uom we can now just multiply (or divide if you want) the uom factor to scale the price
    uom_ids = [
        util.ref(cr, 'openerp_enterprise.uom_categ_online_user'),
        util.ref(cr, 'openerp_enterprise.uom_categ_month_period'),
    ]
    cr.execute(
        """
            UPDATE product_pricing pp
            SET
                price = round(ppli.fixed_price / uom_target.factor, 2)
            FROM
                product_template pt,
                product_product prdct,
                uom_uom uom,
                product_pricelist_item ppli,
                sale_temporal_recurrence str,
                __oe_mig_15_3_recurrence_uom_mapping rum,
                uom_uom uom_target
            WHERE
                pt.id = product_template_id AND
                prdct.product_tmpl_id = pt.id AND
                uom.id = pt.uom_id AND
                (
                    ppli.pricelist_id = pp.pricelist_id AND
                    (ppli.product_tmpl_id = pp.product_template_id OR ppli.product_id = prdct.id)
                ) AND
                (
                    rum.duration = str.duration AND rum.unit = str.unit AND rum.category_id = uom.category_id
                ) AND
                uom_target.id = rum.uom_id AND
                str.id = pp.recurrence_id AND
                uom.category_id = ANY(%s)
        """,
        [uom_ids]
    )

    cr.execute('DROP TABLE __oe_mig_15_3_recurrence_uom_mapping')
